export default class AST_RENDERER {

    constructor(oGlobalTemplates, oGlobalContext, oLocalTemplates, oLocalContext) {

        this.mTemplates = new Map();

        // Will be set when constructed
        this.oData = null;
        this.vTempData = null;

        this.registerTemplates(oGlobalTemplates);

        if (oLocalTemplates) {
            this.registerTemplates(oLocalTemplates, true);
        }

    }

    registerTemplates(oTemplates, bLocal = false) {

        if (oTemplates) {

            for (let sTemplate in oTemplates) {

                if (!this.mTemplates.has(sTemplate)) {

                    this.mTemplates.set(sTemplate, oTemplates[sTemplate]);
                }
                else {

                    console.error(`Attempted to recreate a template: ${sTemplate}`);
                }

            }

            console.log(this.mTemplates);

        }

    }

    convertObjectToString(vObject) {

        let sResolvedValue = null;

        if (Array.isArray(vObject)) {

            sResolvedValue = "[Array]";
        }
        else if (vObject === null) {

            sResolvedValue = "[Null]";
        }
        else {

            sResolvedValue = "[Object]";
        }

        return sResolvedValue;
    }
  
    resolveLogic(oLogic, oData) {

        const F_EXTRACT_EXP_VALUE = (oExpression, oData) => {

            // First evaluated eval
            switch (oExpression.eval.type) {

                case "literal":
                    
                    return oExpression.eval.value;

                case "data":

                    console.log("Expresion resolve data", oData);

                    return this.resolveData(oExpression.eval, oData, false);

                default:

                    console.error("Unknown expression type: ${oExpression.eval.value}");

            }


        }

        const F_RESOLVE_CONDITION = (oCondition, oData) => {

            console.log("oCondition");
            console.log(oCondition);

            console.log("F_RESOLVE_COND", oCondition, oData);

            if (oCondition.type === "unary") {
    
                let vExpressValue = F_EXTRACT_EXP_VALUE(oCondition.expression, oData);

                if (oCondition.expression.inverse) {

                    return (vExpressValue) ? false: true;
                }
                else {

                    return (vExpressValue) ? true : false;
                }

            }
            else if (oCondition.type === "binary") {

                let vLeftEval = F_EXTRACT_EXP_VALUE(oCondition.expression.left, oData);
                let vRightEval = F_EXTRACT_EXP_VALUE(oCondition.expression.right, oData);

                // We will assume the result is false and toggle to true if needed!
                let bResult = false;

                let sOP = null;

                if (typeof oCondition.expression.op === "stirng") {
                    sOP = oCondition.expression.op;
                }
                else if (typeof oCondition.expression.op === "objcet" && oCondition.expression.op.type === "literal") {
                    sOP = oCondition.expression.op.value;
                }

                console.log(sOP);

                switch (oCondition.expression.op) {

                    case "==":

                        if (vLeftEval == vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "===":

                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }
                        break;

                    case "!=":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case "!==":
                        if (vLeftEval === vRightEval) {
                            bResult = true;
                        }  
                        break;

                    case ">":

                        if (vLeftEval > vRightEval) {
                            bResult = true;
                        } 
                        break;
                        
                    case ">=":

                        if (vLeftEval >= vRightEval) {
                            bResult = true;
                        } 
                        break;

                    case "<":

                        if (vLeftEval < vRightEval) {
                            bResult = true;
                        } 
                        break;


                    case "<=":

                        if (vLeftEval <= vRightEval) {
                            bResult = true;
                        } 
                        break;

                }

                return bResult;
            }

        }

        // Loop through all of the supplied conditionals, as 
        for (let oConditional of oLogic.conditionals) {

            if (F_RESOLVE_CONDITION(oConditional.conditional, oData)) {
                return oConditional.consequence
            }

        }

        if (oLogic.alternative && oLogic.alternative.length) {

            if (oLogic.alternative.length === 1 && Object.keys(oLogic.alternative).length === 1 && (oLogic.alternative[0].contents && oLogic.alternative[0].contents.length)) {

                return oLogic.alternative[0].contents;
            }


            return oLogic.alternative;
        }

        return null;
    }

    resolveData(oNode, oData, bError = true) {

        // Continue only if the contents node for data has a string with length
        if (oNode && typeof oNode.value === "string" && oNode.value.length) {

            let asPath = oNode.value.split('.');

            if (asPath[0] === "this") {
                asPath = asPath.slice(1);
            }

            let oContext = null;

            // Determine the "data" context
            switch (asPath[0]) {

                case "_global":

                    oContext = this.oGlobalContext;
                    asPath = asPath.slice(1);
                    break;

                case "_data":

                    oContext = oData;
                    asPath = asPath.slice(1);
                    break;

                default:

                    oContext = oData;
                    break;
            }   

            // Loop through the paths
            for (let p = 0, pLen = asPath.length; p < pLen; p++) {

                // Check to see if this part of the path exists!
                if (oContext.hasOwnProperty(asPath[p])) {

                    if (p + 1 === pLen) {

                        return oContext[asPath[p]];
                    }
                    else {

                        oContext = oContext[asPath[p]];
                        continue;
                    }

                }
                else {

                    if (bError) {
                        console.error(`Requested invalid data context path: ${oNode.value}`);
                    }

                    return null;
                }

            }

        }

    }

    generateAttributes(aoAttributes, oData) {

        let oAttributes = {};

        const F_HANDLE_LITERAL_ATTR = (oLiteralNode, sAttrName) => {

            switch (typeof oLiteralNode.value) {

                case "string":
                    oAttributes[sAttrName].push(oLiteralNode.value);
                    break;

                case "number":
                case "boolean":
                    oAttributes[osAttrName].push(oLiteralNode.value.toString());
                    break;
            }

        }

        const F_HANDLE_DATA_ATTR = (oDataNode, sAttrName, oData) => {

            let vDataValue = this.resolveData(oDataNode, oData);

            switch(typeof vDataValue) {

                case "boolean":
                case "number":
                    vDataValue = vDataValue.toString();
                    break;
                
                case "object":

                    vDataValue = this.convertObjectToString(vDataValue);
                    break;

            }

            if (vDataValue) {

                oAttributes[sAttrName].push(vDataValue);
            }

        }

        const F_LOOP_ATTRIBUTES = (aoAttributes, sSpecificAttr, oData) => {

            // Loop through attribute objects
            for (let oNode of aoAttributes) {
                
                // Shift through all the object types
                switch (oNode.type) {
        
                    // Handle static typed attributes, or those that have dynamics inside of the attribute
                    case "attr":
    
                        if (!oAttributes[oNode.name]) {
                            oAttributes[oNode.name] = [];
                        }
    
                        if (oNode.contents && oNode.contents.length) {
    
                            for (let av = 0, avLen = oNode.contents.length; av < avLen; av++) {
    
                                switch (oNode.contents[av].type) {
    
                                    case "literal":
    
                                        F_HANDLE_LITERAL_ATTR(oNode.contents[av], oNode.name);
                                        break
    
                                    case "data":

                                        
                                        F_HANDLE_DATA_ATTR(oNode.contents[av], oNode.name, oData);
                                        break;
                                        
                                    case "logic":
                                    case "logical":
    
                                        let aoLogicResolve = this.resolveLogic(oNode.contents[av], oData);
    
                                        if (aoLogicResolve && aoLogicResolve.length) {
                                            F_LOOP_ATTRIBUTES(aoLogicResolve, oNode.name, oData);
                                        }
    
                                        break;
    
                                }
    
                            }
    
                        }
    
                        break;

                    case "logic":

                        let aoLogicResolve = this.resolveLogic(oNode, oData);
    
                        if (aoLogicResolve && aoLogicResolve.length) {
                            F_LOOP_ATTRIBUTES(aoLogicResolve, null, oData);
                        }

                        break;

                    default:

                        if (sSpecificAttr) {

                            if (!oAttributes[sSpecificAttr]) {
                                oAttributes[sSpecificAttr] = [];
                            }

                            switch(oNode.type) {

                                case "literal":
            
                                    F_HANDLE_LITERAL_ATTR(oNode, sSpecificAttr);
                                    break;

                                case "data":

                                    F_HANDLE_DATA_ATTR(oNode, sSpecificAttr);
                                    break;
                            }

                        }
    
                }
                
            }

            if (Object.keys(oAttributes).length) {

                return oAttributes;
            }

            return null;
        }

        // Cal the intial attributes list
        F_LOOP_ATTRIBUTES(aoAttributes, null, oData);

        if (Object.keys(oAttributes).length) {

            return oAttributes;
        }

        return null;
    }

    generate(vTemplate = null, oLocalData, dParentNode) {

        let aoTemplate = null;

        if (vTemplate) {

            if (Array.isArray(vTemplate)) {

                aoTemplate = vTemplate;
            }
            else if (typeof vTemplate === "string") {

                if (this.mTemplates.has(vTemplate)) {

                    aoTemplate = this.mTemplates.get(vTemplate);
                }
                else {

                    console.error(`Unknown template was requested: ${vTemplate}`);
                }

            }
            else {

                console.error(`Unexpected data type sent in instead of template name or AST`);
            }

        }
        else {

            console.error(`You must provide a template AST or spcifiy a known template name`);
        }

        // Only continue, if you have a template loaded into memory
        if (aoTemplate) {

            let dRenderedNodes = document.createDocumentFragment();

            for (let a = 0, aLen = aoTemplate.length; a < aLen; a++) {

                let oNode = aoTemplate[a];

                switch (oNode.type) {

                    case "tag":
            
                        // Call the create element function
                        let dNewElement = document.createElement(oNode.name);

                        // Check and resolve attribute objects 
                        if (oNode.attributes && Array.isArray(oNode.attributes) && oNode.attributes.length) {
            
                            let oFinalizedAttributes = this.generateAttributes(oNode.attributes, oLocalData);
            
                            if (oFinalizedAttributes) {
            
                                for (let sAttributeName in oFinalizedAttributes) {
            
                                    switch(sAttributeName) {
            
                                        case "checked":
                                        case "multi":
                                        case "selected":
            
                                            console.warn(`valueless attributes not supported`);
                                            break;
            
                                        default:
            
                                            dNewElement.setAttribute(sAttributeName, oFinalizedAttributes[sAttributeName].join(" ").trim());
                                            break;
            
                                    }
            
                                }
            
                            }
                        }
            
                        // Check and render children
                        if (oNode.contents && Array.isArray(oNode.contents) && oNode.contents.length) {
            
                            let dChildren = this.generate(oNode.contents, oLocalData);

                            if (dChildren) {

                                // Loop through all children and add them to the parent (dNewEelement)
                                while (dChildren.hasChildNodes()) {  
                
                                    dNewElement.appendChild(dChildren.firstChild.cloneNode(true))
                
                                    dChildren.removeChild(dChildren.firstChild);
                                }

                            }
            
                        }
            
                        dRenderedNodes.appendChild(dNewElement);
                        break;

                    case "literal":

                        switch(typeof oNode.value) {

                            case "string":

                                dRenderedNodes.appendChild(document.createTextNode(`${oNode.value}`));
                                break;

                            case "object":

                                dRenderedNodes.appendChild(document.createTextNode(this.convertObjectToString(oNode.vlaue)));
                                break;

                            case "number":
                            case "boolean":

                                dRenderedNodes.appendChild(document.createTextNode(oNode.value.toString()));
                                break;
                        }

                        break;

                    case "data":

                        // In this location, the value is being called after or inside of a tag node

                        let sResolvedValue = this.resolveData(oNode, oLocalData);

                        switch (typeof sResolvedValue) {

                            case "object":

                                sResolvedValue = this.convertObjectToString(sResolvedValue);
                                break;

                            case "boolean":
                            case "number":

                                sResolvedValue = sResolvedValue.toString();
                                break;

                        }

                        dRenderedNodes.appendChild(document.createTextNode(sResolvedValue));

                        break;

                    case "logic":

                        console.log("Logic node");
                        console.log(oNode);

                        // Resolve the logic node
                        let aoLogicResolve = this.resolveLogic(oNode, oLocalData);
                        
                        // If anything is returned, process it with a internal generate
                        if (aoLogicResolve && aoLogicResolve.length) {

                            let dChildren = this.generate(aoLogicResolve, oLocalData);

                            if (dChildren) {

                                // Loop through all children and add them to the parent (dNewEelement)
                                while (dChildren.hasChildNodes()) {  
                
                                    dRenderedNodes.appendChild(dChildren.firstChild.cloneNode(true))
                
                                    dChildren.removeChild(dChildren.firstChild);
                                }

                            }

                        }
                        
                        break;
                }

            }

            return dRenderedNodes;
        }

        return null;

        // for (let a = 0, aLen = aoSource.length; a < aLen; a++) {

        //     let oNode = aoSource[a];

        //     switch (oNode.type) {

        //         case "log":

        //             for (let oContents of oNode.contents) {

        //                 let sLog = null;

        //                 switch (oContents.type) {

        //                     case "literal":

        //                         sLog = oContents.value;
        //                         break;
                                
        //                     case "data":
                            
        //                         sLog = this.resolveData(oNode);
        //                         break;

        //                     default:

        //                         console.error(`[PLATES LOG] => Only supports literal and data value, was passed a ${oNode.type} node.`);
        //                         break;
                                    
        //                 }

        //                 if (sLog !== null) {

        //                     console.log(`[PLATES LOG] => ${oContents.value}`);
        //                 }
        //             }
                            

        //             break;

        //         case "loop":

        //             let vDataSet = this.resolveData(oNode.source);

        //             if (vDataSet !== null) {

        //                 switch (oNode.name) {

        //                     case "for":
    
        //                         if (typeof vDataSet === "object") {

        //                             let iLoopIndex = 0;

        //                             if (Array.isArray(vDataSet)) {
    
        //                                 for (let vData of vDataSet) {
    
        //                                     // Create a new instance, passinf only the templates, and contexts we know about
        //                                     let cLoopSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

        //                                     let oLoopData = {
        //                                         "@index": iLoopIndex,
        //                                         "@key": iLoopIndex,
        //                                         "@source": vDataSet,
        //                                         "this": vData
        //                                     }

        //                                     let dLoopChildren = cLoopSubRender.generate(oNode.contents, false, oLoopData);

        //                                     // Loop through all children and add them to the parent (dNewEelement)
        //                                     while (dLoopChildren.hasChildNodes()) {  

        //                                         dParentNode.appendChild(dLoopChildren.firstChild.cloneNode(true))

        //                                         dLoopChildren.removeChild(dLoopChildren.firstChild);
        //                                     }

                                            
        //                                     // Null out the template subrenderer
        //                                     cLoopSubRender = null;
                                            
        //                                     iLoopIndex += 1;
        //                                 }
    
        //                             }
        //                             else {
    
        //                                 for (let sDatasetKey in vDataSet) {

        //                                     // Create a new instance, passinf only the templates, and contexts we know about
        //                                     let cLoopSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

        //                                     let oLoopData = {
        //                                         "@index": iLoopIndex,
        //                                         "@key": sDatasetKey,
        //                                         "@source": vDataSet,
        //                                         "this": vDataSet[sDatasetKey],
        //                                     }

        //                                     console.log("Object loop");

        //                                     console.log(oLoopData);

        //                                     let dLoopChildren = cLoopSubRender.generate(oNode.contents, false, oLoopData);

        //                                     // Loop through all children and add them to the parent (dNewEelement)
        //                                     while (dLoopChildren.hasChildNodes()) {  

        //                                         dParentNode.appendChild(dLoopChildren.firstChild.cloneNode(true))

        //                                         dLoopChildren.removeChild(dLoopChildren.firstChild);
        //                                     }

                                            
        //                                     // Null out the template subrenderer
        //                                     cLoopSubRender = null;
                                            
        //                                     iLoopIndex += 1;

        //                                 }
                                        
        //                             }
    
        //                         }
        //                         else {
    
        //                             console.error(`Data source is not an interable: ${vDataSet}`);
        //                         }
    
        //                         break;
    
        //                 }

        //             }

        //             break;

        //         case "template":

        //             // verify this template is real first
        //             if (this.mTemplates.has(oNode.name)) {

        //                 // Create a new instance, passinf only the templates, and contexts we know about
        //                 let cSubRender = new AST_RENDERER(this.oTemplates, this.oGlobalContext, this.oLocalContext);

        //                 if (oNode.arguments && oNode.arguments.length) {

        //                     let oNewData = {};

        //                     for (let arg of oNode.arguments) {

        //                         if (typeof arg === "object") {

        //                             switch (arg.type) {

        //                                 case "data":

        //                                     let vData = this.resolveData(arg);

        //                                     if (typeof vData === "object" && !Array.isArray(vData)) {

        //                                         oNewData = {
        //                                             ...oNewData,
        //                                             ...vData
        //                                         };

        //                                     }
        //                                     else {

        //                                         console.log(`Template arguments can only be references to another object or a statically asigned key value pair where the value can be any data type.`)
        //                                     }

        //                                     break;

        //                                 case "keyValue":

        //                                     let sKey = arg.key;
        //                                     let vKeyValue = null;

        //                                     for (let oStaticContents of arg.contents) {

        //                                         switch(oStaticContents.type) {

        //                                             case "literal":
                                                    
        //                                                 if (vKeyValue) {
        //                                                     vKeyValue += oStaticContents.value;
        //                                                 }
        //                                                 else {

        //                                                     vKeyValue = oStaticContents.value;
        //                                                 }

        //                                                 break;

        //                                             case "data":

        //                                                 let vData = this.resolveData(oStaticContents);

        //                                                 if (vKeyValue) {

        //                                                     vKeyValue += vData;
        //                                                 }
        //                                                 else {

        //                                                     vKeyValue = vData;
        //                                                 }

        //                                                 break;

        //                                         }
                                                
        //                                     }

        //                                     oNewData[sKey] = vKeyValue;
        //                                     break;
        //                             }

        //                         }

        //                     }

        //                     let dSubChildren = cSubRender.generate(this.mTemplates.get(oNode.name), false, oNewData);

        //                     // Loop through all children and add them to the parent (dNewEelement)
        //                     while (dSubChildren.hasChildNodes()) {  

        //                         dParentNode.appendChild(dSubChildren.firstChild.cloneNode(true))

        //                         dSubChildren.removeChild(dSubChildren.firstChild);
        //                     }

                            
        //                     // Null out the template subrenderer
        //                     cSubRender = null;
        //                 }
        //                 else {

        //                     this.generate(this.mTemplates.get(oNode.name), dParentNode);
        //                 }
                        
        //             }
        //             else {

        //                 console.error(`Requested unknown template: ${oNode.name}`);
        //             }

        //             break;

        //         default:

        //             console.error(`Unknown tag was discovered: ${oNode.type}`);
        //             break;
        //     }

        // }

        // if (this.vTempData !== null) {
        //     this.vTempData = null;
        // }

        // return dParentNode;

    }
}